/*
 * (c) Copyright 1992 by Panagiotis Tsirigotis
 * (c) Sections Copyright 1998-2001 by Rob Braun
 * All rights reserved.  The file named COPYRIGHT specifies the terms 
 * and conditions for redistribution.
 */

#ifndef SCONST_H
#define SCONST_H

/*
 * $Id$
 */

/*
 * Names of internal non-visible services
 */
#define INTERCEPT_SERVICE_NAME         "intercept"
#define LOG_SERVICE_NAME               "logging"


/*
 * Log entry ids
 */
#define START_ENTRY                    "START"
#define FAIL_ENTRY                     "FAIL"
#define EXIT_ENTRY                     "EXIT"
#define USERID_ENTRY                   "USERID"
#define NOID_ENTRY                     "NOID"

#endif    /* SCONST_H */

